@@ -36,7 +36,7 @@ import io
 import re
 import socket
 import time
-from datetime import datetime
+from datetime import datetime, timedelta
 from uuid import UUID
 import warnings
 
@@ -565,11 +565,16 @@ class DateType(_CassandraType):
 
     @staticmethod
     def deserialize(byts, protocol_version):
-        return datetime.utcfromtimestamp(int64_unpack(byts) / 1000.0)
+        timestamp = int64_unpack(byts) / 1000.0
+        if timestamp >= 0:
+            dt = datetime.utcfromtimestamp(timestamp)
+        else:
+            # PYTHON-119: workaround for Windows
+            dt = datetime(1970, 1, 1) + timedelta(seconds=timestamp)
+        return dt
 
     @staticmethod
     def serialize(v, protocol_version):
-        global _have_warned_about_timestamps
         try:
             converted = calendar.timegm(v.utctimetuple())
             converted = converted * 1e3 + getattr(v, 'microsecond', 0) / 1e3
@@ -578,6 +583,7 @@ class DateType(_CassandraType):
             if type(v) not in _number_types:
                 raise TypeError('DateType arguments must be a datetime or timestamp')
 
+            global _have_warned_about_timestamps
             if not _have_warned_about_timestamps:
                 _have_warned_about_timestamps = True
                 warnings.warn(
